#include <GL/glew.h>
#include "view3d.h"
#include "mesh.h"
#include <QGLFormat>
#include <QMouseEvent>
#include "utils.h"
#include <QDebug>

namespace
{
	const QGLFormat gl_format(QGL::DoubleBuffer | QGL::Rgba | QGL::DepthBuffer | QGL::DirectRendering);
}

View3d::View3d(QWidget *parent) :
	QGLWidget(gl_format, parent)
{
	mesh = new Mesh;
	center = Vec3(0,0,0);
	radius = 1.f;

	makeCurrent();

	glewInit();
}

View3d::~View3d()
{
	makeCurrent();
	delete mesh;
}

void View3d::initializeGL()
{
	glViewport(0,0,width(),height());

	glClearColor(0.75,0.75,1.0,0);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);

	glDisable(GL_CULL_FACE);
	glDisable(GL_BLEND);
	glEnable(GL_NORMALIZE);

	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
	glEnable(GL_LIGHTING);

	const GLfloat ambient[] = { 0.f, 0.f, 0.f, 0.f };
	const GLfloat diffuse[] = { 1.f, 1.f, 1.f, 1.f };
	const GLfloat specular[] = { 1.f, 1.f, 1.f, 1.f };
	const GLfloat position[] = { 1.f / std::sqrt(3.f), 1.f / std::sqrt(3.f), 1.f / std::sqrt(3.f), 0.f };
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular);
	glLightfv(GL_LIGHT0, GL_POSITION, position);
	glEnable(GL_LIGHT0);

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glScalef(1.f,-1.f,1.f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void View3d::resizeGL(int w, int h)
{
	glViewport(0,0,w,h);
}

void View3d::paintGL()
{
	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, double(width()) / height(), 0.75 * radius, 2.75 * radius);
	glMatrixMode(GL_MODELVIEW);

	glPushMatrix();

	glTranslatef(center[0],center[1],center[2]);
	glMultMatrixd((const GLdouble*)mesh_mat.data());
	glTranslatef(-center[0],-center[1],-center[2]);

	mesh->draw();

	glPopMatrix();
}

void View3d::load(const QString &filename)
{
	makeCurrent();

	mesh->load(filename);
	mesh->optimize();

	const std::pair<Vec3, Vec3> &bbox = mesh->getBoundingBox();
	center = 0.5f * (bbox.first + bbox.second);
	radius = 0.5f * (bbox.second - bbox.first).length();

	const Vec3 view_dir(0,0,-1);
	const Vec3 up(0,1,0);
	const Vec3 eye = center - 1.75f * radius * view_dir;

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(eye[0], eye[1], eye[2], center[0], center[1], center[2], up[0], up[1], up[2]);

	mesh_mat = QMatrix4x4();

	update();
}

void View3d::save(const QString &filename) const
{
	mesh->save(filename);
}

void View3d::mouseMoveEvent(QMouseEvent *e)
{
	if (!(e->buttons() & Qt::LeftButton))
		return;

	if (last_cursor_pos == e->posF())
		return;

	const Vec3 view_dir(0,0,-1);
	const Vec3 up(0,1,0);
	const Vec3 right(1,0,0);
	const Vec3 eye = center - 3.0f * radius * view_dir;

	const float fy = tan(30.0 * 0.5 * M_PI / 180.0) / height() * 2.0;
	const float fx = fy * height() / width();
	Vec3 dir = float(last_cursor_pos.x() - 0.5 * width()) * fx * right
			   - float(last_cursor_pos.y() - 0.5 * height()) * fy * up
			   + view_dir;
	dir.normalize();
	last_cursor_pos = e->posF();

	Vec3 hp0;
	if (!hit_sphere(eye, dir, center, radius, hp0))
		return;

	dir = float(last_cursor_pos.x() - 0.5 * width()) * fx * right
		  - float(last_cursor_pos.y() - 0.5 * height()) * fy * up
		  + view_dir;
	dir.normalize();

	Vec3 hp1;
	if (!hit_sphere(eye, dir, center, radius, hp1))
		return;

	hp0 -= center;
	hp0.normalize();
	hp1 -= center;
	hp1.normalize();

	Vec3 w = hp0 ^ hp1;
	const float angle = asin(w.length()) * 180.0f / M_PI;
	w.normalize();

	QMatrix4x4 mat;
	mat.rotate(angle, w[0], w[1], w[2]);
	mesh_mat = mat * mesh_mat;

	update();
}

void View3d::mousePressEvent(QMouseEvent *e)
{
	if (e->button() == Qt::LeftButton)
		last_cursor_pos = e->posF();
}
